"""
:copyright: (c) 2017 by Jumper Labs Ltd.
:license: Apache 2.0, see LICENSE.txt for more details.
"""
from __future__ import print_function

import os
import sys
from time import sleep
import calendar
import time
import requests
import logging

API_URL = 'https://us-central1-jemu-web-app.cloudfunctions.net/api/v1' if 'JUMPER_STAGING' not in os.environ else \
    'https://us-central1-jemu-web-app-staging.cloudfunctions.net/api/v1'


class WebException(Exception):
    pass


class AuthorizationError(WebException):
    def __init__(self, message):
        super(WebException, self).__init__(message)
        self.exit_code = 4
        self.message = message


class UnInitializedError(WebException):
    def __init__(self):
        print ("Failed to get user id. Please reach out to support@jumper.io for help")
        super(WebException, self).__init__("Failed to get user id. Please reach out to support@jumper.io for help")
        self.exit_code = 6
        self.message = "Failed to get user id. Please reach out to support@jumper.io for help"


class EmulatorGenerationError(WebException):
    def __init__(self, message):
        super(WebException, self).__init__(message)
        self.exit_code = 5
        self.message = message


class JemuWebApi(object):
    def __init__(self, jumper_token=None, api_url=API_URL):
        self._api_url = api_url
        self._token = jumper_token
        self._headers = {'Authorization': 'Bearer ' + self._token}
        self._user_uid = None
        self.init()

    def init(self):
        sys.stdout.write('Loading virtual device\n')
        sys.stdout.flush()
        logging.getLogger("requests").setLevel(logging.WARNING)
        res = requests.get(self._api_url + '/hello', headers=self._headers)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == requests.codes['unauthorized'] or res.status_code == requests.codes['forbidden']:
                print ("Error: Authorization failed. Check the token in your config.json file")
                raise AuthorizationError("Error: Authorization failed. Check the token in the config.json file.")
            else:
                raise e

        self._user_uid = res.json()['userUid']

    def upload_file(self, filename, data):
        if self._user_uid is None:
            raise UnInitializedError

        headers = self._headers
        headers['Content-Type'] = 'application/octet-stream'
        res = requests.post(
            '{}/firmwares/{}/{}?zip={}'.format(self._api_url, self._user_uid, filename, 'true'),
            data=data,
            headers=headers
        )
        res.raise_for_status()

    def check_status(self, filename):
        if self._user_uid is None:
            raise UnInitializedError

        headers = self._headers
        headers['Content-Type'] = 'application/text'
        res = requests.get(
            '{}/firmwares/{}/{}/status'.format(self._api_url, self._user_uid, filename),
            headers=headers
        )
        return res

    def send_event(self, event_name):
        if self._user_uid is None:
            raise UnInitializedError

        headers = self._headers
        headers['Content-Type'] = 'application/text'
        res = requests.post(
            '{}/analytics/{}/{}'.format(self._api_url, self._user_uid, event_name),
            data='',
            headers=headers
        )
        return res

    def check_error(self, filename):
        if self._user_uid is None:
            raise UnInitializedError

        headers = self._headers
        headers['Content-Type'] = 'application/text'
        res = requests.get(
            '{}/firmwares/{}/{}/error'.format(self._api_url, self._user_uid, filename),
            headers=headers
        )
        return res

    def get_jemu_version(self):
        headers = self._headers
        headers['Content-Type'] = 'application/text'
        res = requests.get(
            '{}/jemu-version'.format(self._api_url),
            headers=headers
        )
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            return None
        return res.text

    def download_jemu(self, filename, local_filename):
        if self._user_uid is None:
            raise UnInitializedError

        headers = self._headers
        headers['Content-Type'] = 'application/octet-stream'
        res = requests.get(
            '{}/firmwares/{}/{}'.format(self._api_url, self._user_uid, filename),
            headers=headers)
        
        res = requests.get(res.text)

        with open(local_filename, 'wb') as f:
            for chunk in res.iter_content(chunk_size=1024): 
                if chunk: # filter out keep-alive new chunks
                    f.write(chunk)
                    
        return True

    def get_archived_emulator(self, fw_filename, fw_bin_data, dest):
        signs = {"$", "#", "[", "]"}
        num_of_signs = fw_filename.count('.') - 1
        fw_filename = fw_filename.replace('.', '_', num_of_signs)
        for i in signs:
            fw_filename = fw_filename.replace(i, '_')

        fw_filename = str(int(calendar.timegm(time.gmtime()))) + '_' + fw_filename
        self.upload_file(fw_filename, fw_bin_data)

        sys.stdout.flush()
        status = 'Queded'
        while status != 'Done':
            status = self.check_status(fw_filename).text
            sys.stdout.write('.')
            sys.stdout.flush()
            sleep(0.25)
            if status == 'Fail':
                self.send_event('firmware processing failed')
                sys.stdout.write("FAILED")
                sys.stdout.write('\nVirtual device failed to start. Please reach out to support@jumper.io for help\n')
                sys.stdout.flush()
                error_string = self.check_error(fw_filename).text
                raise EmulatorGenerationError(error_string)

        self.send_event('firmware processing sucess')
        jemu_filename = os.path.splitext(fw_filename)[0] + '.jemu'

        self.download_jemu(jemu_filename, dest)
        self.send_event('firmware download success')

        sys.stdout.write(' Done\n')
        sys.stdout.flush()



"""
:copyright: (c) 2017 by Jumper Labs Ltd.
:license: Apache 2.0, see LICENSE.txt for more details.
"""
from queue import Queue
from jemu_peripheral import JemuPeripheral


class JemuMemPeripheral(JemuPeripheral):
    _LEVEL_HIGH = 1
    _LEVEL_LOW = 0
    _RESPONSE_TIMEOUT = 10

    _TYPE_STRING = "type"
    _VALUE = "value"
    _PERIPHERAL_ID = "peripheral_id"
    _PERIPHERAL_TYPE = "peripheral_type"
    _REG_ADDRESS = "address"
    _COMMAND = "command"
    _COMMAND_SET_REG = "set_reg_value"
    _COMMAND_GET_REG = "get_reg_value"

    def __init__(self, jemu_connection, id, peripheral_type, generators):
        JemuPeripheral.__init__(self, jemu_connection, id, peripheral_type, generators)
        self._queue = Queue()
        self._jemu_connection.register(self._receive_callback)

    def _set_reg_json(self, register, value):
        return {
            self._REG_ADDRESS: register,
            self._TYPE_STRING: self._COMMAND,
            self._VALUE: value,
            self._PERIPHERAL_ID: self._id,
            self._COMMAND: self._COMMAND_SET_REG,
            self._PERIPHERAL_TYPE: self._peripheral_type
        }

    def _get_reg_json(self, register):
        return {
            self._REG_ADDRESS: register,
            self._TYPE_STRING: self._COMMAND,
            self._PERIPHERAL_ID: self._id,
            self._COMMAND: self._COMMAND_GET_REG,
            self._PERIPHERAL_TYPE: self._peripheral_type
        }

    def _receive_callback(self, jemu_packet):
        if 'type' in jemu_packet and jemu_packet["type"] == "value_response":
            self._got_message = True
            self._queue.put(jemu_packet["value"])

    def set_register_value(self, register, value):
        self._jemu_connection.send_json(self._set_reg_json(register, value))

    def get_register_value(self, register):
        self._jemu_connection.send_json(self._get_reg_json(register))
        return int(self._queue.get(timeout=self._RESPONSE_TIMEOUT))

    def status(self):
        return self._jemu_connection.recv_json()



"""
:copyright: (c) 2017 by Jumper Labs Ltd.
:license: Apache 2.0, see LICENSE.txt for more details.
"""
import os
import json
from .jemu_button import JemuButton
from .jemu_mem_peripheral import JemuMemPeripheral
from .jemu_bma280 import JemuBMA280
from .jemu_bme280 import JemuBME280
from .jemu_sudo import JemuSudo
from .jemu_bq24160 import JemuBQ24160
from .jemu_bq27421 import JemuBQ27421


class JemuPeripheralsParser:
    _EXTERNAL_PERIPHERAL = "External"
    _NRF52_PERIPHERAL = "nrf52"

    __peripherals_json_path = None

    def __init__(self, peripherals_json_path):
        self.__peripherals_json_path = peripherals_json_path

    def get_peripherals(self, jemu_connection):
        peripheral_list = []
        mcu_list = []

        if not os.path.isfile(self.__peripherals_json_path):
            raise Exception(self.__peripherals_json_path + ' is not found')
        elif not os.access(self.__peripherals_json_path, os.R_OK):
            raise Exception(self.__peripherals_json_path + ' is not readable')
        else:
            with open(self.__peripherals_json_path) as peripherals_json_file:
                peripherals_json = json.load(peripherals_json_file)

                for peripheral in peripherals_json["Peripherals"]:
                    peripheral_id = peripheral["id"]

                    if ("config" in peripheral) and ("generators" in peripheral["config"]):
                        generators = peripheral["config"]["generators"]
                    else:
                        generators = None

                    if peripheral["class"] == "Button":
                        peripheral_obj = JemuButton(jemu_connection, peripheral_id)
                        peripheral_list.append({"obj": peripheral_obj, "name": peripheral["name"]})

                    if peripheral["class"] == "BME280":
                        peripheral_obj = JemuMemPeripheral(jemu_connection, peripheral_id, self._EXTERNAL_PERIPHERAL, generators)
                        peripheral_list.append({"obj": peripheral_obj, "name": peripheral["name"]})

                    if peripheral["class"] == "M24C04":
                        peripheral_obj = JemuMemPeripheral(jemu_connection, peripheral_id, self._EXTERNAL_PERIPHERAL, generators)
                        peripheral_list.append({"obj": peripheral_obj, "name": peripheral["name"]})

                    if peripheral["class"] == "AD7798":
                        peripheral_obj = JemuMemPeripheral(jemu_connection, peripheral_id, self._EXTERNAL_PERIPHERAL, generators)
                        peripheral_list.append({"obj": peripheral_obj, "name": peripheral["name"]})

                    if peripheral["class"] == "ADG1404":
                        peripheral_obj = JemuMemPeripheral(jemu_connection, peripheral_id, self._EXTERNAL_PERIPHERAL, generators)
                        peripheral_list.append({"obj": peripheral_obj, "name": peripheral["name"]})

                    if peripheral["class"] == "PCA9539A":
                        peripheral_obj = JemuMemPeripheral(jemu_connection, peripheral_id, self._EXTERNAL_PERIPHERAL, generators)
                        peripheral_list.append({"obj": peripheral_obj, "name": peripheral["name"]})

                    if peripheral["class"] == "BQ24160":
                        peripheral_obj = JemuBQ24160(jemu_connection, peripheral_id, self._EXTERNAL_PERIPHERAL, generators)
                        peripheral_list.append({"obj": peripheral_obj, "name": peripheral["name"]})

                    if peripheral["class"] == "BQ27421":
                        peripheral_obj = JemuBQ27421(jemu_connection, peripheral_id, self._EXTERNAL_PERIPHERAL, generators)
                        peripheral_list.append({"obj": peripheral_obj, "name": peripheral["name"]})
                    
                    if peripheral["class"] == "BMA280":
                        peripheral_obj = JemuBMA280(jemu_connection, peripheral_id, self._EXTERNAL_PERIPHERAL, generators)
                        peripheral_list.append({"obj": peripheral_obj, "name": peripheral["name"]})

                    if peripheral["class"] == "BME280":
                        peripheral_obj = JemuBME280(jemu_connection, peripheral_id, self._EXTERNAL_PERIPHERAL, generators)
                        peripheral_list.append({"obj": peripheral_obj, "name": peripheral["name"]})

                    if peripheral["class"] == "MCU":
                        mcu_list.append(peripheral["name"])

                if "InternalsPeripherals" not in peripherals_json:
                    return peripheral_list

                # Load internak peripherals
                internal_peripherals = peripherals_json["InternalsPeripherals"]
                for mcu in mcu_list:
                    if mcu not in internal_peripherals:
                        continue
                    peripherals = internal_peripherals[mcu]
                    for peripheral in peripherals:
                        if peripheral["class"] == "SUDO":
                            peripheral_obj = JemuSudo(jemu_connection, peripheral["id"], mcu)
                            peripheral_list.append({"obj": peripheral_obj, "name": peripheral["class"]})
                
            return peripheral_list



"""
:copyright: (c) 2017 by Jumper Labs Ltd.
:license: Apache 2.0, see LICENSE.txt for more details.
"""
from __future__ import print_function

import unittest

from jemu.jemu_web_api import JemuWebApi, UnInitializedError


class TestWebApi(unittest.TestCase):
    def setUp(self):
        self.api = JemuWebApi()

    def tearDown(self):
        pass

    def test_init(self):
        self.assertIsNotNone(self.api._user_uid)

    def test_uninit_upload_file(self):
        self.api._user_uid = None
        with self.assertRaises(UnInitializedError):
            self.api.upload_file('test_file', b'SOME_BINARY_DATA')

    def test_upload_file(self):
        self.api.upload_file('test_file.bin', b'SOME_BINARY_DATA')



"""
:copyright: (c) 2017 by Jumper Labs Ltd.
:license: Apache 2.0, see LICENSE.txt for more details.
"""
from __future__ import print_function
import argparse
import signal
from vlab import VlabException
from jemu_web_api import WebException
from os import getcwd
from time import sleep
from threading import Event
from __version__ import __version__ as jumper_current_version

import sys

from .vlab import Vlab
from .vlab_hci_device import VirtualHciDevice
from . import __version__


class _VersionAction(argparse.Action):
    def __init__(self,
                 option_strings,
                 version=None,
                 dest=argparse.SUPPRESS,
                 default=argparse.SUPPRESS,
                 help="show program's version number and exit"):
        super(_VersionAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)
        self.version = version

    def __call__(self, parser, namespace, values, option_string=None):
        version = self.version
        if version is None:
            version = parser.version
        formatter = parser._get_formatter()
        formatter.add_text(version)
        Vlab.check_version()
        parser.exit(message=formatter.format_help())


def run(args):
    try:
        vlab = Vlab(
            working_directory=args.directory,
            sudo_mode=args.sudo,
            gdb_mode=args.gdb,
            print_trace=args.trace,
            trace_output_file=args.trace_dest,
            print_uart=args.uart,
            uart_output_file=args.uart_dest
        )

        vlab.load(args.bin)

        reached_bkpt = Event()

        def bkpt_callback(code):
            print('Firmware reached a BKPT instruction with code {}'.format(code))
            reached_bkpt.set()

        vlab.on_bkpt(bkpt_callback)

        def pins_listener(pin_number, pin_level):
            print("pin_number:", pin_number, "  pin_level:", pin_level)

        if args.gpio:
            print("gpio")
            vlab.on_pin_level_event(pins_listener)


        vlab.start()

        if (not args.uart) and (not args.trace):
            print('\nVirtual device is running without UART/Trace prints (use -u and/or -t to get your firmware execution status)')
        else:
            print('\nVirtual device is running')

        while (not reached_bkpt.is_set()) and vlab.is_running():
            sleep(0.1)

        if reached_bkpt and vlab.is_running():
            vlab.stop()

        if vlab.get_return_code():
            sys.exit(vlab.get_return_code())

    except VlabException as e:
        print(e.message)
        exit(e.exit_code)


def hci():
    virtual_hci_device = VirtualHciDevice()
    virtual_hci_device.start()
    try:
        while True:
            sleep(0.1)
    except KeyboardInterrupt:
        pass
    finally:
        virtual_hci_device.stop()


def signal_handler(signal, frame):
    sys.exit(1)


def main():
    signal.signal(signal.SIGINT, signal_handler)
    parser = argparse.ArgumentParser(
        prog='jumper',
        description="CLI interface for using Jumper's emulator"
    )

    parser.add_argument('--version', action=_VersionAction, version='%(prog)s {}'.format(__version__))

    subparsers = parser.add_subparsers(title='Commands', dest='command')

    run_parser = subparsers.add_parser(
        'run',
        help='Runs an emulator with a binary FW file. Currently only support nRF52 devices'
    )
    run_parser.add_argument(
        '--bin',
        '-b ',
        help='Binary file used to generate the FW from',
    )
    run_parser.add_argument(
        '--directory',
        '-d ',
        help='Working directory, must include the peripherals.json and scenario.json files. Default is current working directory',
        default=getcwd()
    )

    run_parser.add_argument(
        '--sudo',
        '-s ',
        help='Run in sudo mode => FW can write to read-only registers. This should usually be used for testing low-level drivers',
        action='store_true',
        default=False
    )

    run_parser.add_argument(
        '--gdb',
        '-g ',
        help='Opens a GDB port for debugging the FW on port 5555. The FW will not start running until the GDB client connects.',
        action='store_true',
        default=False
    )

    run_parser.add_argument(
        '--version',
        '-v ',
        help='Jumper sdk version.',
        action='store_true',
        default=False
    )

    run_parser.add_argument(
        '--trace',
        '-t ',
        action='store_true',
        help='Prints a trace report to stdout, this can be used with --trace-dest to forward it to a file.',
        default=False
    )

    run_parser.add_argument(
        '--trace-dest',
        type=str,
        help=
        """
        Forwards the trace report to a destination file. This MUST be used -t with this flag to make it work.
        To print to stdout, just hit -t.
        """,
        default="",
    )

    run_parser.add_argument(
        '--uart',
        '-u ',
        action='store_true',
        default=False,
        help='Forwards UART prints to stdout, this can be used with --uart-dest to forward it to a file.'
    )

    run_parser.add_argument(
        '--uart-dest',
        type=str,
        help=
        """
        Forwards UART prints to a destination file. This MUST be used -u with this flag to make it work.
        To print to stdout, just hit -u.
        """,
        default="",
    )

    run_parser.add_argument(
        "--gpio",
        help=
        """
        Prints GPIO events to stdout.
        """,
        action='store_true',
        default=False,
    )

    run_parser = subparsers.add_parser(
        'ble',
        help='Creates a virtual HCI device (BLE dongle) for regular Linux/Bluez programs to communicate with virtual devices'
    )

    args = parser.parse_args()

    if args.command == 'run':
        if args.version:
            print ("v" + jumper_current_version)
            exit(0)
        if not args.bin:
            print("jumper run: error: argument --bin/-b is required")
            exit(1)
        run(args)

    if args.command == 'ble':
        hci()


if __name__ == '__main__':
    main()
